"""Conv module.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Union

import torch
from torch import nn

from kindle.modules.activation import Activation
from kindle.utils.torch_utils import autopad


class Conv(nn.Module):
    """Standard convolution with batch normalization and activation.

    Arguments: [channel, kernel_size, stride, padding, groups, activation]
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        padding: Union[int, None] = None,
        groups: int = 1,
        activation: Union[str, None] = "ReLU",
    ) -> None:
        """Standard convolution with batch normalization and activation.

        Args:
            in_channels: input channels.
            out_channels: output channels.
            kernel_size: kernel size.
            stride: stride.
            padding: input padding. If None is given, autopad is applied
                which is identical to padding='SAME' in TensorFlow.
            groups: group convolution.
            activation: activation name. If None is given, nn.Identity is applied
                which is no activation.
        """
        super().__init__()
        # error: Argument "padding" to "Conv2d" has incompatible type "Union[int, List[int]]";
        # expected "Union[int, Tuple[int, int]]"
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding=autopad((kernel_size, kernel_size), padding),  # type: ignore
            groups=groups,
            bias=False,
        )
        self.batch_norm = nn.BatchNorm2d(out_channels)
        self.activation = Activation(activation)()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return self.activation(self.batch_norm(self.conv(x)))

    def fuseforward(self, x: torch.Tensor) -> torch.Tensor:
        """Fuse forward."""
        return self.activation(self.conv(x))


class Focus(Conv):
    """Focus module from YOLOv5.

    Arguments: [channel, kernel_size, stride, padding, groups, activation]
    """

    @classmethod
    def __reduce_focus(cls, x: torch.Tensor) -> torch.Tensor:
        """Run reduce image by half."""

        return torch.cat(
            [
                x[..., ::2, ::2],
                x[..., 1::2, ::2],
                x[..., ::2, 1::2],
                x[..., 1::2, 1::2],
            ],
            1,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Reduce image size and run convolution."""
        return super().forward(self.__reduce_focus(x))

    def fuseforward(self, x: torch.Tensor) -> torch.Tensor:
        """Reduce image size and run convolution with batchnorm fused."""
        return super().fuseforward(self.__reduce_focus(x))



"""Module Description.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import Any, Dict, List

import numpy as np
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract
from kindle.modules.add import Add


class AddGenerator(GeneratorAbstract):
    """Add module generator."""

    def __init__(self, *args, **kwargs) -> None:
        """Add module generator."""
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> int:
        if isinstance(self.from_idx, int):
            raise Exception("Add must have more than 2 inputs.")

        return self.in_channels[self.from_idx[0]]

    @property
    def in_channel(self) -> int:
        return self.out_channel

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self._get_kwargs(Add, self.args)

    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        return list(size)

    def __call__(self, repeat: int = 1) -> nn.Module:
        module = Add()

        return self._get_module(module)



"""Pretrained generator module.

Generate pre-trained module that is served from https://github.com/rwightman/pytorch-image-models

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import Any, Dict, List, Union

import numpy as np
import timm
import torch
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract
from kindle.modules.pretrained import Pretrained, PretrainedFeatureMap


class PreTrainedGenerator(GeneratorAbstract):
    """Generate Pretrained module."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> Union[int, List[int]]:
        temp_in_shape = np.array([self.in_channel, 128, 128])
        out_shape = self.compute_out_shape(temp_in_shape)

        if isinstance(out_shape[0], list):
            return [o_shape[0] for o_shape in out_shape if isinstance(o_shape, list)]

        return out_shape[0]

    @property
    def in_channel(self) -> int:
        return self.in_channels[self.from_idx]  # type: ignore

    def compute_out_shape(
        self, size: np.ndarray, repeat: int = 1
    ) -> Union[List[int], List[List[int]]]:
        module = self(repeat=repeat)
        module_out = module(torch.zeros([1, *list(size)]))

        if isinstance(module_out, torch.Tensor):
            return list(module_out.shape[1:])

        return [list(m_out.shape[1:]) for m_out in module_out]

    @property
    def kwargs(self) -> Dict[str, Any]:
        kwargs = self._get_kwargs(Pretrained, self.args)
        return kwargs

    def __call__(self, repeat: int = 1) -> nn.Module:
        module = Pretrained(**self.kwargs)
        return self._get_module(module)


class PreTrainedFeatureMapGenerator(GeneratorAbstract):
    """Generate PreTrainedFeatureMap."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> int:
        return self.in_channels[self.from_idx][self.kwargs["feature_idx"]]  # type: ignore

    @property
    def in_channel(self) -> int:
        return self.in_channels[self.from_idx][self.kwargs["feature_idx"]]  # type: ignore

    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        return list(size[self.kwargs["feature_idx"]])

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self._get_kwargs(PretrainedFeatureMap, self.args)

    def __call__(self, repeat: int = 1) -> nn.Module:
        module = PretrainedFeatureMap(**self.kwargs)
        return self._get_module(module)


if __name__ == "__main__":
    m = timm.create_model("mobilenetv3_large_100", pretrained=True, features_only=True)



"""Module Description.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import List, Tuple, Union

import torch
from torch import nn


class Add(nn.Module):
    """Add module for Kindle."""

    def __init__(self):  # pylint: disable=useless-super-delegation
        """Initialize module."""
        super().__init__()

    @classmethod
    def forward(
        cls, x: Union[Tuple[torch.Tensor, ...], List[torch.Tensor]]
    ) -> torch.Tensor:
        """Add inputs.

        Args:
            x: list of torch tensors

        Returns:
            sum of all x's
        """
        result = x[0]
        for i in range(1, len(x)):
            result = result + x[i]

        return result



"""Module Description.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from copy import deepcopy
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import torch
import yaml
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract
from kindle.model import Model


def convert_yaml_args(
    data: List, args: Union[List, Tuple], inplace: bool = True
) -> List:
    """Convert yaml data with argument value.

    Args:
        data: list or tuple that might contain string of 'arg0', 'arg1' ...
        args: argument values to replace with 'arg0', 'arg1' ...
        inplace: if False, it will not overwrite value.

    Returns:
        overwritten values by replacing 'arg0', 'arg1' ... to args[0], args[1] ...
    """
    if not inplace:
        data = deepcopy(data)

    for i, _ in enumerate(data):
        if isinstance(data[i], list):
            data[i] = convert_yaml_args(data[i], args)
        elif isinstance(data[i], str) and data[i].startswith("arg"):
            idx = int(data[i][3:])
            data[i] = args[idx]

    return data


class YamlModuleGenerator(GeneratorAbstract):
    """Custom yaml module generator."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize YamlModuleGenerator."""
        super().__init__(*args, **kwargs)
        with open(self.args[0], "r", encoding="utf-8") as f:
            self.cfg = yaml.load(f, yaml.SafeLoader)

        for i in range(1, len(self.args)):
            self.cfg["args"][i - 1] = self.args[i]

        if "args" in self.cfg and len(self.args) > 1:
            convert_yaml_args(self.cfg["module"], self.cfg["args"])

        self.cfg.update(
            {
                "input_channel": self.in_channel,
                "depth_multiple": 1.0,
                "width_multiple": self.width_multiply,
                "backbone": self.cfg.pop("module"),
            }
        )

    @property
    def out_channel(self) -> int:
        temp_in_shape = np.array([self.cfg["input_channel"], 128, 128])
        out_shape = self.compute_out_shape(temp_in_shape)
        return out_shape[0]

    @property
    def in_channel(self) -> int:
        if isinstance(self.from_idx, int):
            return self.in_channels[self.from_idx]
        return sum([self.in_channels[idx] for idx in self.from_idx])

    @torch.no_grad()
    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        module = self(repeat=repeat)
        module.eval()
        module_out = module(torch.zeros([1, *list(size)]))

        return list(module_out.shape[-3:])

    @property
    def kwargs(self) -> Dict[str, Any]:
        kwargs = {"verbose": False}
        if self.keyword_args is not None and "verbose" in self.keyword_args:
            kwargs["verbose"] = self.keyword_args["verbose"]
        return kwargs

    def __call__(self, repeat: int = 1) -> nn.Module:
        module: Union[List[nn.Module], nn.Module]

        if repeat > 1:
            # Currently, yaml module must have same in and out channel in order to apply repeat.
            module = [Model(self.cfg, **self.kwargs) for _ in range(repeat)]
        else:
            module = Model(self.cfg, **self.kwargs)

        return self._get_module(module)



"""PyTorch trainer module.

** This module is experimental and tend to be removed and changed in the future. **

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Optional, Tuple, Union

import numpy as np
import torch
from torch import nn, optim
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.sampler import SequentialSampler, SubsetRandomSampler
from tqdm import tqdm


def _get_n_data_from_dataloader(dataloader: DataLoader) -> int:
    """Get a number of data in dataloader.

    Args:
        dataloader: torch data loader

    Returns:
        A number of data in dataloader
    """
    if isinstance(dataloader.sampler, SubsetRandomSampler):
        n_data = len(dataloader.sampler.indices)
    elif isinstance(dataloader.sampler, SequentialSampler):
        n_data = len(dataloader.sampler.data_source)
    else:
        n_data = len(dataloader) * dataloader.batch_size if dataloader.batch_size else 1

    return n_data


def _get_n_batch_from_dataloader(dataloader: DataLoader) -> int:
    """Get a batch number in dataloader.

    Args:
        dataloader: torch data loader

    Returns:
        A batch number in dataloader
    """
    n_data = _get_n_data_from_dataloader(dataloader)
    n_batch = dataloader.batch_size if dataloader.batch_size else 1

    return n_data // n_batch


class TorchTrainer:
    """Pytorch Trainer."""

    def __init__(
        self,
        model: nn.Module,
        criterion: nn.Module,
        optimizer: optim.Optimizer,
        device: Union[str, torch.device] = "cpu",
        verbose: int = 1,
    ) -> None:
        """Initialize TorchTrainer class.

        Args:
            model: model to train
            criterion: loss function module
            optimizer: optimization module
            device: torch device
            verbose: verbosity level.
        """

        if isinstance(device, str):
            device = torch.device(device)

        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.verbose = verbose
        self.device = device

    def train(
        self,
        train_dataloader: DataLoader,
        n_epoch: int,
        shuffle: bool = False,
        test_dataloader: Optional[DataLoader] = None,
    ) -> Tuple[float, float]:
        """Train model.

        Args:
            train_dataloader: data loader module which is a iterator that returns (data, labels)
            n_epoch: number of total epochs for training
            test_dataloader: test data loader
            shuffle: shuffle train data on every epoch.
                     Sampler must be SubsetRandomSampler to apply shuffle.

        Returns:
            loss and accuracy
        """
        average_loss, accuracy = -1.0, -1.0
        n_batch = _get_n_batch_from_dataloader(train_dataloader)

        for epoch in range(n_epoch):
            if shuffle and isinstance(train_dataloader.sampler, SubsetRandomSampler):
                np.random.shuffle(train_dataloader.sampler.indices)

            running_loss = 0.0
            correct = 0
            total = 0

            pbar = tqdm(enumerate(train_dataloader), total=n_batch)
            for batch, (data, labels) in pbar:
                data, labels = data.to(self.device), labels.to(self.device)
                self.optimizer.zero_grad()

                model_out = self.model(data)
                loss = self.criterion(model_out, labels)
                loss.backward()
                self.optimizer.step()

                # TODO: Modify for multi-label classification.
                _, predicted = torch.max(model_out, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

                running_loss += loss.item()
                pbar.update()
                pbar.set_description(
                    f"Train: [{epoch + 1:03d}] "
                    f"Loss: {(running_loss / (batch + 1)):.7f}, "
                    f"Accuracy: {(correct / total) * 100:.2f}%"
                )
            pbar.close()

            if test_dataloader is not None:
                self.test(test_dataloader)

            average_loss = running_loss / n_batch
            accuracy = correct / total

        return average_loss, accuracy

    @torch.no_grad()
    def test(self, test_dataloader: DataLoader) -> Tuple[float, float]:
        """Test model.

        Args:
            test_dataloader: test data loader module which is a iterator that returns (data, labels)

        Returns:
            loss, accuracy
        """

        n_batch = _get_n_batch_from_dataloader(test_dataloader)

        running_loss = 0.0
        correct = 0
        total = 0

        pbar = tqdm(enumerate(test_dataloader), total=n_batch)
        for batch, (data, labels) in pbar:
            data, labels = data.to(self.device), labels.to(self.device)
            model_out = self.model(data)
            running_loss += self.criterion(model_out, labels).item()

            # TODO: Modify for multi-label classification.
            _, predicted = torch.max(model_out, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()

            pbar.update()
            pbar.set_description(
                f" Test: {'':5} Loss: {(running_loss / (batch + 1)):.7f}, "
                f"Accuracy: {(correct / total) * 100:.2f}%"
            )

        loss = running_loss / n_batch
        accuracy = correct / total
        return loss, accuracy



"""UpSample module generator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Any, Dict, List

import numpy as np
import torch
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract


class UpSampleGenerator(GeneratorAbstract):
    """UpSample module generator."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> int:
        return self.in_channel

    @property
    def in_channel(self) -> int:
        return self.in_channels[self.from_idx]  # type: ignore

    @property
    def kwargs(self) -> Dict[str, Any]:
        kwargs = self._get_kwargs(nn.Upsample, self.args)
        if kwargs["scale_factor"] is None:
            kwargs["scale_factor"] = 2  # Default scale factor is 2

        return kwargs

    @torch.no_grad()
    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        module = self(repeat=repeat)
        module_out = module(torch.zeros([1, *list(size)]))

        return list(module_out.shape[1:])

    def __call__(self, repeat: int = 1) -> nn.Module:
        return self._get_module(nn.Upsample(**self.kwargs))



"""Common utility functions.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

import math
from typing import List, Tuple, Union

import numpy as np
import torch
from torch import nn
from torch.utils.data.sampler import SubsetRandomSampler


def split_dataset_index(
    n_data: int, split_ratio: float = 0.1
) -> Tuple[SubsetRandomSampler, SubsetRandomSampler]:
    """Split dataset indices with split_ratio.

    Args:
        n_data: number of total data
        split_ratio: split ratio (0.0 ~ 1.0)

    Returns:
        SubsetRandomSampler ({split_ratio} ~ 1.0)
        SubsetRandomSampler (0 ~ {split_ratio})
    """
    indices = np.arange(n_data)
    split = int(split_ratio * indices.shape[0])

    train_idx = indices[split:]
    valid_idx = indices[:split]

    train_sampler = SubsetRandomSampler(train_idx)
    valid_sampler = SubsetRandomSampler(valid_idx)

    return train_sampler, valid_sampler


def make_divisible(n_channel: Union[int, float], divisor: int = 8) -> int:
    """Convert {n_channel} to divisible by {divisor}

    Args:
        n_channel: number of channels.
        divisor: divisor to be used.

    Returns:
        Ex) n_channel=22, divisor=8
            ceil(22/8) * 8 = 24
    """
    return int(math.ceil(n_channel / divisor) * divisor)


def autopad(
    kernel_size: Union[int, List[int]], padding: Union[int, None] = None
) -> Union[int, List[int]]:
    """Auto padding calculation for pad='same' in TensorFlow."""
    # Pad to 'same'
    if isinstance(kernel_size, int):
        kernel_size = [kernel_size]

    return padding or [x // 2 for x in kernel_size]


def count_model_params(
    model: torch.nn.Module,
) -> int:
    """Count model's parameters."""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def fuse_conv_and_batch_norm(conv: nn.Conv2d, batch_norm: nn.BatchNorm2d) -> nn.Conv2d:
    """Fuse convolution and batchnorm layers.

    https://tehnokv.com/posts/fusing-batchnorm-and-conv/

    Args:
        conv: convolution module.
        batch_norm: Batch normalization module directly connected to the conv module.

    Return:
        Fused conv with batch norm.
    """

    fused_conv = (
        nn.Conv2d(
            conv.in_channels,
            conv.out_channels,
            kernel_size=conv.kernel_size,  # type: ignore
            stride=conv.stride,  # type: ignore
            padding=conv.padding,  # type: ignore
            groups=conv.groups,
            bias=True,
        )
        .requires_grad_(False)
        .to(conv.weight.device)
    )

    # Fusing weight
    w_conv = conv.weight.clone().view(conv.out_channels, -1)
    w_batch_norm = torch.diag(
        batch_norm.weight.div(
            torch.sqrt(batch_norm.eps + batch_norm.running_var)  # type: ignore
        )
    )
    fused_conv.weight.copy_(
        torch.mm(w_batch_norm, w_conv).view(fused_conv.weight.size())
    )

    # Fusing bias
    if conv.bias is None:
        b_conv = torch.zeros(conv.weight.size(0), device=conv.weight.device)
    else:
        b_conv = conv.bias

    b_batch_norm = batch_norm.bias - batch_norm.weight.mul(
        batch_norm.running_mean  # type: ignore
    ).div(
        torch.sqrt(batch_norm.running_var + batch_norm.eps)  # type: ignore
    )
    fused_conv.bias.copy_(  # type: ignore
        torch.mm(w_batch_norm, b_conv.reshape(-1, 1)).reshape(-1) + b_batch_norm
    )

    return fused_conv



"""Module Description.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
import ast
from typing import Union

from torch import nn


class Activation:
    """Convert string activation name to the activation class."""

    def __init__(self, act_type: Union[str, None]) -> None:
        """Convert string activation name to the activation class.

        Args:
            act_type: Activation name.

        Returns:
            nn.Identity if {type} is None.
        """
        self.name = act_type
        self.args = [1] if self.name == "Softmax" else []

    def __call__(self) -> nn.Module:
        if self.name is None:
            return nn.Identity()
        if hasattr(nn, self.name):
            return getattr(nn, self.name)(*self.args)

        return ast.literal_eval(self.name)()



"""YOLOv5 head module.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

import math
from typing import List, Optional, Tuple, Union

import torch
from torch import nn


class YOLOHead(nn.Module):
    """YOLOv5 head module."""

    export: bool = False

    def __init__(  # pylint: disable=not-callable
        self,
        n_classes: int,
        anchors: Tuple[Tuple[float, ...]],
        n_channels: Tuple[int, ...],
        strides: Tuple[float, ...],
        out_xyxy: bool = False,
    ) -> None:
        """Initialize YOLOHead.

        Args:
            n_classes: number of classes.
            anchors: anchor arrays. (n_layer, n_anchor*2).
            out_xyxy: return coordinates as xyxy format (For older yolov5 compatability)
            n_channels: number of in channels from the feature map.
            strides: stride sizes from the feature map.
        """
        super().__init__()

        self.out_xyxy = out_xyxy
        # Number of classes
        self.n_classes = n_classes
        # Number of outputs per anchor
        self.n_outputs = n_classes + 5
        # Number of detection layers
        self.n_layers = len(anchors)
        # Number of anchors per a layer
        self.n_anchors = len(anchors[0]) // 2

        # YOLOv5 compatability
        self.nc = self.n_classes  # pylint: disable=invalid-name
        self.no = self.n_outputs  # pylint: disable=invalid-name
        self.nl = self.n_layers  # pylint: disable=invalid-name
        self.na = self.n_anchors  # pylint: disable=invalid-name

        self.grid = [
            torch.zeros(1) for _ in range(self.n_layers)
        ]  # Grid initialization

        anchor_buf = torch.tensor(anchors).float().view(self.n_layers, -1, 2)
        self.register_buffer("anchors", anchor_buf)  # (n_layer, n_anchors, 2)
        self.register_buffer(
            "anchor_grid", anchor_buf.clone().view(self.n_layers, 1, -1, 1, 1, 2)
        )  # (n_layers, 1, n_anchors, 1, 1, 2)

        # TODO(jeikeilim): Conv type can be choosable.
        self.conv = nn.ModuleList(
            nn.Conv2d(x, self.n_outputs * self.n_anchors, 1) for x in n_channels
        )
        self.stride = torch.tensor(strides)
        self.anchors /= self.stride.view(-1, 1, 1)  # pylint: disable=no-member

        # TODO(jeikeilim): Consider what to do with NMS layer

    def forward(
        self, x: List[torch.Tensor]
    ) -> Union[
        Tuple[torch.Tensor, List[torch.Tensor]], Tuple[torch.Tensor], List[torch.Tensor]
    ]:
        """Forward YOLO head.

        Args:
            x: List of feature maps

        Returns:
            1. (batch_size, n_anchors, height, width, n_classes + 5) in training mode.
            2. (batch_size, n_anchors * heights * widths, n_classes + 5) in export mode
            3. (1) + (2) in inference mode
        """
        preds = []
        for i in range(self.n_layers):
            x[i] = self.conv[i](x[i])
            batch_size, _, height, width = x[i].shape

            # Reshape to (batch_size, n_anchor, height, width, outputs)
            x[i] = (
                x[i]
                .view(batch_size, self.n_anchors, self.n_outputs, height, width)
                .permute(0, 1, 3, 4, 2)
                .contiguous()
            )

            if not self.training:
                if self.grid[i].shape[2:4] != x[i].shape[2:4]:
                    self.grid[i] = self._make_grid(width, height).to(x[i].device)
                else:
                    self.grid[i] = self.grid[i].to(x[i].device)

                y = x[i].sigmoid()
                box_xy = (y[..., 0:2] * 2.0 - 0.5 + self.grid[i]) * self.stride[i]
                box_wh = (y[..., 2:4] * 2) ** 2 * self.anchor_grid[i]  # type: ignore

                if self.out_xyxy:
                    box_xyxy = self._xywh2xyxy(box_xy, box_wh).view(batch_size, -1, 4)
                    y = torch.cat(
                        (box_xyxy, y[..., 4:].view(batch_size, -1, self.n_outputs - 4)),
                        -1,
                    )
                else:
                    box_xy = box_xy.view(batch_size, -1, 2)
                    box_wh = box_wh.view(batch_size, -1, 2)
                    score = y[..., 4:].view(batch_size, -1, self.n_classes + 1)
                    y = torch.cat((box_xy, box_wh, score), -1)

                preds.append(y)

        if self.training:
            return x
        if self.export:
            return (torch.cat(preds, 1),)

        return (torch.cat(preds, 1), x)

    def initialize_biases(
        self,
        class_frequency: Optional[torch.Tensor] = None,
        n_object_per_image: Tuple[int, int] = (8, 640),
        class_probability: float = 0.6,
    ) -> None:
        """Initialize biases for objectness and class probabilities.

        Args:
        class_frequency: Class histogram values. If this is provided,
            class biases will be initialized by
            initial bias + log(class_frequency / class_frequency.sum())

        n_object_per_image: Number of object per image. (Number of object, per image).
            objectness biases will be initialized by
            initial bias + log(n_object / (per_image / stride) ** 2)

        class_probability: Global class probability.
            This will be ignored if class_frequency is provided.
            Otherwise, class biasese will be initialized by
            initial bias + log(class_probability / (n_classes - 0.99))
        """

        for conv, stride in zip(self.conv, self.stride):
            bias = conv.bias.view(self.n_anchors, -1)
            bias.data[:, 4] += math.log(
                n_object_per_image[0] / (n_object_per_image[1] / stride) ** 2
            )

            if class_frequency is None:
                bias.data[:, 5:] += math.log(
                    class_probability / (self.n_classes - 0.99)
                )
            else:
                bias.data[:, 5:] += torch.log(class_frequency / class_frequency.sum())

            conv.bias = nn.Parameter(bias.view(-1), requires_grad=True)

    @staticmethod
    def _make_grid(width: int = 20, height: int = 20) -> torch.Tensor:
        """Make grid."""
        y_grid, x_grid = torch.meshgrid([torch.arange(height), torch.arange(width)])
        return torch.stack((x_grid, y_grid), 2).view((1, 1, height, width, 2)).float()

    @staticmethod
    def _xywh2xyxy(  # pylint: disable=invalid-name
        xy: torch.Tensor, wh: torch.Tensor
    ) -> torch.Tensor:
        """Convert [x, y, w, h] to [x1, y1, x2, y2]."""
        # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2]
        # where xy1=top-left, xy2=bottom-right
        half_wh = wh / 2
        x1y1 = xy - half_wh
        x2y2 = xy + half_wh
        return torch.cat([x1y1, x2y2], -1)



"""Kindle Model parser and model.

This module parses model configuration yaml file
and generates PyTorch model accordingly.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import torch
import yaml
from torch import nn

from kindle.generator.base_generator import GeneratorAbstract, ModuleGenerator
from kindle.generator.flatten import FlattenGenerator
from kindle.generator.yolo_head import YOLOHeadGenerator
from kindle.modules import Conv, DWConv, Focus, YOLOHead
from kindle.utils.model_utils import ModelInfoLogger, ModelProfiler
from kindle.utils.torch_utils import fuse_conv_and_batch_norm


class Model(nn.Module):
    """PyTorch model class."""

    def __init__(
        self,
        cfg: Union[str, Dict[str, type]],
        verbose: bool = False,
    ) -> None:
        """Parse model from the model config file.

        Args:
            cfg: yaml file path or dictionary type of the model.
            verbose: print the model parsing information.
        """
        super().__init__()
        self.model_parser = ModelParser(cfg=cfg, verbose=verbose)
        self.model = self.model_parser.model
        self.output_save = self.model_parser.output_save

    def forward(
        self,
        x: Union[torch.Tensor, np.ndarray],
        augment_func: Optional[Union[List[Callable], Callable]] = None,
        n_augment: int = 3,
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        """Forward the model.

        Args:
            x: model input. When augment_func is given,
                {X} can be either torch.Tensor or np.ndarray.
                However, output of augmentation function must emit torch.Tensor.

            augment_func: augmentation function.
                        If list of functions is given, TTA will be applied by original input {X} and
                        {X} with each augmentation functions.
                        Output shape will be (len({AUGMENT_FUNC})+1, batch_size, output_size).

                        If single function is given, TTA will be applied by original input {X} and
                        {X} {AUGMENT_FUNC} applied by {N_AUGMENT} times.
                        Output shape will be ({N_AUGMENT}+1, batch_size, output_size)

            n_augment: number of augmentations to apply on TTA augmentation.
                       Only applies when single augmentation function is given.

        Returns:
            Model output.
        """
        if augment_func is not None:
            if isinstance(augment_func, list):
                funcs = augment_func
            else:
                funcs = [augment_func for _ in range(n_augment)]

            y_list = [self.forward_once(aug_func(x)) for aug_func in funcs]
            if isinstance(x, torch.Tensor):
                y_list.append(self.forward_once(x))

            y = torch.stack(y_list)
        else:
            assert isinstance(x, torch.Tensor), "Input must be torch tensor for non-TTA"
            y = self.forward_once(x)

        return y

    def fuse(self, verbose: bool = False) -> nn.Module:
        """Fuse Conv - BatchNorm2d layers."""

        if verbose:
            print("Fusing layers ", end="")

        for module in self.model.modules():
            dot_str = "."

            if isinstance(module, (Conv, DWConv, Focus)) and hasattr(
                module, "batch_norm"
            ):
                module.conv = fuse_conv_and_batch_norm(module.conv, module.batch_norm)
                delattr(module, "batch_norm")
                module.forward = module.fuseforward  # type: ignore

                dot_str = ","

            if verbose:
                print(dot_str, end="", flush=True)

        if verbose:
            print(" Done!")

        return self

    def profile(self, verbose: bool = True, **kwargs) -> ModelProfiler:
        """Run model profiler.

        Args:
            verbose: print profiling result.

        Returns:
            ModelProfiler instance which contains profiling result.
        """
        profiler = ModelProfiler(self, **kwargs)
        profiler.run(verbose=verbose)

        return profiler

    def forward_once(
        self, x: torch.Tensor, profile_func: Optional[Callable] = None
    ) -> torch.Tensor:
        """Forward one time only."""
        y: List[Union[torch.Tensor, None]] = []

        for i, module in enumerate(self.model):  # type: ignore
            if module.from_idx != -1:
                x = (
                    y[module.from_idx]  # type: ignore
                    if isinstance(module.from_idx, int)
                    else [x if j == -1 else y[j] for j in module.from_idx]
                )
            if profile_func is not None:
                x = profile_func(module, x, i)
            else:
                x = module(x)
            y.append(x if module.module_idx in self.output_save else None)

        return x


class YOLOModel(Model):
    """YOLO model."""

    def __init__(
        self,
        cfg: Union[str, Dict[str, type]],
        verbose: bool = False,
        init_bias=False,
        **kwargs,
    ) -> None:
        """Parse YOLOModel from the model config file.

        Args:
            cfg: yaml file path or dictionary type of the model.
            verbose: print the model parsing information.
            init_bias: Initialize bias.
            kwargs: Keyword arguments for the YOLOHead.initialize_biases
        """
        super().__init__(cfg, verbose=verbose)

        assert isinstance(self.model[-1], YOLOHead), (
            "YOLOHead must have YOLOHead at the end! "
            f"Current last layer: {self.model[-1].name} "
        )

        self._yolo_init()
        self.initialize_biases = self.model[-1].initialize_biases

        if init_bias:
            self.initialize_biases(**kwargs)

        # YOLOv5 compatability
        self.stride = self.model[-1].stride

    def _yolo_init(self) -> None:
        """Initialize model for YOLO training."""

        # Initialize batch norm
        for module in self.model.modules():
            module_type = type(module)

            if module_type is nn.Conv2d:
                pass
            elif module_type is nn.BatchNorm2d:
                module.eps = 1e-3  # type: ignore
                module.momentum = 0.03  # type: ignore
            elif module_type in (nn.LeakyReLU, nn.ReLU, nn.ReLU6):
                module.inplace = True  # type: ignore


class ModelParser:
    """Generate PyTorch model from the model yaml file."""

    def __init__(
        self,
        cfg: Union[str, Dict[str, Type]] = "./model_configs/show_case.yaml",
        verbose: bool = False,
    ) -> None:
        """Generate PyTorch model from the model yaml file.

        Args:
            cfg: model config file or dict values read from the model config file.
            verbose: print the parsed model information.
        """

        self.verbose = verbose
        if isinstance(cfg, dict):
            self.cfg = cfg
        else:
            with open(cfg, encoding="utf-8") as f:
                self.cfg = yaml.load(f, Loader=yaml.FullLoader)

        self.input_size = None

        if (
            self.cfg
            and "input_size" in self.cfg
            and len(self.cfg["input_size"]) == 2  # type: ignore
        ):
            self.input_size = self.cfg["input_size"]

        self.custom_module_paths: Optional[Union[List[str], str]]
        if "custom_module_paths" in self.cfg:
            self.custom_module_paths = self.cfg["custom_module_paths"]  # type: ignore
        else:
            self.custom_module_paths = None

        self.in_channel = self.cfg["input_channel"]

        self.depth_multiply = self.cfg["depth_multiple"]
        self.width_multiply = self.cfg["width_multiple"]

        self.backbone_cfg: List[List] = self.cfg["backbone"]  # type: ignore
        if "head" in self.cfg:
            self.head_cfg: Optional[List[List]] = self.cfg["head"]  # type: ignore
        else:
            self.head_cfg = None

        if "channel_divisor" in self.cfg:
            self.channel_divisor: int = self.cfg["channel_divisor"]  # type: ignore
        else:
            self.channel_divisor = GeneratorAbstract.CHANNEL_DIVISOR

        self.model, self.output_save = self._parse_model()

    def log(self, msg: str):
        """Log."""
        if self.verbose:
            print(msg)

    def _parse_model(  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        self,
    ) -> Tuple[nn.Sequential, List[int]]:
        """Parse model."""
        in_channels: List[int] = []
        in_sizes: List[int] = []
        layers: List[nn.Module] = []
        output_save: List[int] = []
        logger = ModelInfoLogger(self.input_size is not None)

        if self.head_cfg is not None:
            model_cfg = self.backbone_cfg + self.head_cfg
        else:
            model_cfg = self.backbone_cfg

        channel_divisor = GeneratorAbstract.CHANNEL_DIVISOR
        GeneratorAbstract.CHANNEL_DIVISOR = self.channel_divisor
        for i, module_cfg in enumerate(model_cfg):  # type: ignore
            if len(module_cfg) > 4:
                idx, repeat, module, args, kwargs = module_cfg
            else:
                idx, repeat, module, args = module_cfg
                kwargs = None

            if i >= len(self.backbone_cfg):
                GeneratorAbstract.CHANNEL_DIVISOR = 1
                width_multiply = 1.0
                depth_multiply = 1.0
            else:
                width_multiply = float(self.width_multiply)  # type: ignore
                depth_multiply = float(self.depth_multiply)  # type: ignore

            module_generator = ModuleGenerator(
                module, custom_module_paths=self.custom_module_paths
            )(
                *args,
                keyword_args=kwargs,
                from_idx=idx,
                in_channels=tuple(in_channels) if i > 0 else (self.in_channel,),  # type: ignore
                width_multiply=width_multiply,
            )
            repeat = max(round(repeat * depth_multiply), 1) if repeat > 1 else repeat

            if isinstance(module_generator, FlattenGenerator):
                if self.input_size is not None:
                    module_generator.in_shape = in_sizes[idx]  # type: ignore
                else:
                    module_generator.in_shape = [in_channels[idx], 1, 1]  # type: ignore
            elif isinstance(module_generator, YOLOHeadGenerator):
                module_generator.in_shape = [in_sizes[idx_shape] for idx_shape in idx]
                module_generator.input_size = self.input_size  # type: ignore

            module = module_generator(repeat=repeat)
            module.module_idx, module.from_idx = i, idx

            if self.input_size is not None:
                in_size = (
                    np.array(in_sizes, dtype=np.object_)[idx]
                    if i > 0
                    else [self.in_channel] + self.input_size
                )
                out_size = module_generator.compute_out_shape(in_size, repeat=repeat)
                in_sizes.append(out_size)
            else:
                in_size, out_size = None, None

            in_channels.append(module_generator.out_channel)
            layers.append(module)

            output_save.extend(
                [x % i for x in ([idx] if isinstance(idx, int) else idx) if x != -1]
            )

            logger.add(
                (i, idx, repeat),
                module,
                module_generator,
                args,
                kwargs=kwargs,
                in_size=in_size,
                out_size=out_size,
            )

        GeneratorAbstract.CHANNEL_DIVISOR = channel_divisor
        parsed_model = nn.Sequential(*layers)
        n_param = sum([x.numel() for x in parsed_model.parameters()])
        n_grad = sum([x.numel() for x in parsed_model.parameters() if x.requires_grad])

        self.log(logger.info)
        self.log(
            f"Model Summary: {len(list(parsed_model.modules())):,d} "
            f"layers, {n_param:,d} parameters, {n_grad:,d} gradients"
            f"\n"
        )

        return parsed_model, output_save



"""Concat module generator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import Any, Dict, List

import numpy as np

from kindle.generator.base_generator import GeneratorAbstract
from kindle.modules.concat import Concat


class ConcatGenerator(GeneratorAbstract):
    """Concatenation module generator."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> int:
        """Get out channel size."""
        # error: Value of type "Optional[List[int]]" is not indexable
        # error: Item "int" of "Union[int, List[int]]" has no attribute
        # "__iter__" (not iterable)
        return sum([self.in_channels[i] for i in self.from_idx])  # type: ignore

    @property
    def in_channel(self) -> int:
        """Get in channel size."""
        return -1

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self._get_kwargs(Concat, self.args)

    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        """Compute out shape."""
        return [self.out_channel] + list(size[0][1:])

    def __call__(self, repeat: int = 1):
        module = Concat(**self.kwargs)

        return self._get_module(module)



"""Pretrained module.

Pre-trained model downloaded from https://github.com/rwightman/pytorch-image-models
Please refer to https://rwightman.github.io/pytorch-image-models/ for the supported model overview.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import List, Union

import timm
import torch
from torch import nn


class Pretrained(nn.Module):
    """Pretrained module.

    Please refer to https://rwightman.github.io/pytorch-image-models/ for the supported
    model overview.
    """

    def __init__(
        self,
        model_name: str,
        use_feature_maps: bool = False,
        features_only: bool = True,
        pretrained: bool = True,
    ) -> None:
        """Initialize Pretrained instance.

        Args:
            model_name: name of the model from timm.list_models(pretrained=True)
            use_feature_maps: if True, return value of the module will be list of each feature maps
                            (features_only must be True in this case).
                            Otherwise, returns last feature map.
            features_only: if True, skip classification layer and use feature layers only.
            pretrained: use pretrained weight.
        """
        super().__init__()

        self.pass_last = not use_feature_maps and features_only
        self.module = timm.create_model(
            model_name=model_name, pretrained=pretrained, features_only=features_only
        )

    def forward(self, x: torch.Tensor) -> Union[torch.Tensor, List[torch.Tensor]]:
        """Forward pretrained network.

        Returns:
            if use_feature_maps is True, List of torch.Tensor will be returned.
            which each array contains corresponding feature map of pooling layer.
            Otherwise, torch.Tensor will be returned which is output of the last layer.
        """

        y = self.module(x)

        return y[-1] if self.pass_last else y


class PretrainedFeatureMap(nn.Module):
    """Placeholder of the output of Pretrained module when features_only = True"""

    def __init__(self, feature_idx: int = -1) -> None:
        """Initialize PretrainedFeatureMap.

        Args:
            feature_idx: index of the feature maps.
        """
        super().__init__()
        self.feature_idx = feature_idx

    def forward(self, x: List[torch.Tensor]) -> torch.Tensor:
        """Bypass input of x[self.feature_idx]."""
        return x[self.feature_idx]



"""Kindle is an easy model build package for PyTorch.

Building a deep learning model became so simple that
almost all model can be made by copy and paste from
other existing model codes.

So why code? when we can simply build a model with yaml markup file.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
- URL: https://github.com/JeiKeiLim/kindle
"""

from kindle import generator, modules
from kindle.model import Model, YOLOModel
from kindle.trainer import TorchTrainer
from kindle.utils import ModelProfiler

from .version import __version__

__all__ = [
    "modules",
    "generator",
    "Model",
    "YOLOModel",
    "TorchTrainer",
    "ModelProfiler",
    "__version__",
]



"""Bottleneck(ResNet) module.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from typing import Union

import torch
from torch import nn

from kindle.modules.activation import Activation
from kindle.modules.conv import Conv


class Bottleneck(nn.Module):
    """Standard bottleneck block.

    Arguments: [channel, shortcut, groups, expansion, activation]
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        shortcut=True,
        groups: int = 1,
        expansion: float = 0.5,
        activation: Union[str, None] = "ReLU",
    ) -> None:
        """Initialize Bottleneck instance.

        Args:
            in_channels: number of incoming channels
            out_channels: number of outgoing channels
            shortcut: whether to use shortcut connection.
                This only works when in_channels and out_channels are identical.
            groups: number of group convolution number.
            expansion: expansion ratio.
            activation: Name of the activation to use in convolution.
        """
        super().__init__()
        expansion_channel = int(out_channels * expansion)

        self.conv1 = Conv(in_channels, expansion_channel, 1, 1, activation=activation)
        self.conv2 = Conv(
            expansion_channel, out_channels, 3, 1, groups=groups, activation=activation
        )
        self.shortcut = shortcut and in_channels == out_channels

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        out = self.conv2(self.conv1(x))

        if self.shortcut:
            out = out + x

        return out


class BottleneckCSP(nn.Module):
    """CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks.

    Arguments: [channel, shortcut, groups, expansion, activation]
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        n_repeat: int = 1,
        shortcut=True,
        groups: int = 1,
        expansion: float = 0.5,
        activation: Union[str, None] = "ReLU",
    ) -> None:
        """Initialize BottleneckCSP instance.

        Args:
            in_channels: number of incoming channels
            out_channels: number of outgoing channels
            n_repeat: repeat number of bottleneck.
            shortcut: whether to use shortcut connection.
                This only works when in_channels and out_channels are identical.
            groups: number of group convolution number.
            expansion: expansion ratio.
            activation: Name of the activation to use in convolution.
        """
        super().__init__()

        expansion_channel = int(out_channels * expansion)

        self.conv1 = Conv(in_channels, expansion_channel, 1, 1, activation=activation)
        self.conv2 = nn.Conv2d(in_channels, expansion_channel, 1, 1, bias=False)
        self.conv3 = nn.Conv2d(expansion_channel, expansion_channel, 1, 1, bias=False)
        self.conv4 = Conv(
            2 * expansion_channel, out_channels, 1, 1, activation=activation
        )
        self.batch_norm = nn.BatchNorm2d(2 * expansion_channel)
        self.activation = Activation(activation)()

        self.bottleneck_csp = nn.Sequential(
            *[
                Bottleneck(
                    expansion_channel,
                    expansion_channel,
                    shortcut=shortcut,
                    groups=groups,
                    expansion=1.0,
                    activation=activation,
                )
                for _ in range(n_repeat)
            ]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward bottleneck CSP.

        Args:
            x: input tensor.
        """
        conv_y1 = self.conv3(self.bottleneck_csp(self.conv1(x)))
        conv_y2 = self.conv2(x)

        return self.conv4(
            self.activation(self.batch_norm(torch.cat((conv_y1, conv_y2), dim=1)))
        )


class C3(nn.Module):
    """CSP Bottleneck with 3 convolutions."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        n_repeat: int = 1,
        shortcut=True,
        groups: int = 1,
        expansion: float = 0.5,
        activation: Union[str, None] = "ReLU",
    ) -> None:
        """Initialize BottleneckCSP instance.

        Args:
            in_channels: number of incoming channels
            out_channels: number of outgoing channels
            n_repeat: repeat number of bottleneck.
            shortcut: whether to use shortcut connection.
                This only works when in_channels and out_channels are identical.
            groups: number of group convolution number.
            expansion: expansion ratio.
            activation: Name of the activation to use in convolution.
        """
        super().__init__()
        expansion_channel = int(out_channels * expansion)

        self.conv1 = Conv(in_channels, expansion_channel, 1, 1, activation=activation)
        self.conv2 = Conv(in_channels, expansion_channel, 1, 1, activation=activation)
        self.conv3 = Conv(
            2 * expansion_channel, out_channels, 1, activation=activation
        )  # act=FReLU(out_channels)

        self.bottleneck_c3 = nn.Sequential(
            *[
                Bottleneck(
                    expansion_channel,
                    expansion_channel,
                    shortcut=shortcut,
                    groups=groups,
                    expansion=1.0,
                    activation=activation,
                )
                for _ in range(n_repeat)
            ]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward Bottleneck C3.

        Args:
            x: input tensor
        """
        return self.conv3(
            torch.cat((self.bottleneck_c3(self.conv1(x)), self.conv2(x)), dim=1)
        )



"""PyTorch Modules."""

from kindle.modules.activation import Activation
from kindle.modules.add import Add
from kindle.modules.bottleneck import C3, Bottleneck, BottleneckCSP
from kindle.modules.concat import Concat
from kindle.modules.conv import Conv, Focus
from kindle.modules.dwconv import DWConv
from kindle.modules.linear import Linear
from kindle.modules.poolings import SPP, GlobalAvgPool
from kindle.modules.yolo_head import YOLOHead

__all__ = [
    "Add",
    "Bottleneck",
    "BottleneckCSP",
    "C3",
    "Concat",
    "Conv",
    "DWConv",
    "Focus",
    "Linear",
    "GlobalAvgPool",
    "SPP",
    "Activation",
    "YOLOHead",
]
